/*
 * Startup Code for ARM
 *
 * Assembler file for SWI
 * this file handles the very basic aspects of
 * the SWI-Instruction.
 *
 * by: Clemens Helfmeier
 * ClemensHelfmeier {at] gmx `dot" de
 * (replace the corresponding at and dot in the email, please).
 */
#define ASSEMBLER

#include "arm.inc"
#include "swi_asm.h"
#include "startup_config.h"

.text
.func SWI_Handler
.global SWI_Handler
SWI_Handler:
#if (ENABLE_SWI == 1)
	/* the SoftwareInterrupt doesn't need to save
	 * all registers since it doesn't use all of them
	 * always. The application could also save the registers
	 * because it knows exactly where the SWI happens.
	 */
	/* the SWI is executed in Supervisor mode. Take care
	 * of what it does.
	 */
	/* some SWI codes are needed in IRQs etc. thus they should not take too long
	 * they are implemented in ASM allowing very fast responses.
	 * currently these are: SWI_IRQ,
	 */

	/* since SWI does not call itself, it doesn't need to save SPSR. */
	stmfd	sp!, {r0,r1}

	/* load the requested SWI Code from the instruction we came from*/
	ldr	r0, [lr, #-4]
	ldr	r1, =0xFF
	bic	r0, r0, r1, LSL#24
	/* determine the requested function */
	ldr	r1, =SWI_IRQ
	cmp	r0, r1
	bne	SWI_Handler_a
		/* SWI_IRQ. if from within IRQ-Routine, set IRQ-Mode and
		 * turn off IRQ-Interrupts, leave FIQs as they are and return */
		.extern IRQ_Handler
		.extern IRQ_Handler_Size
		ldr	r0, =IRQ_Handler
		sub	r0, lr, r0
		sub	r0, r0, #4
		ldr	r1, =(IRQ_Handler_Size-4)
		cmp	r0, r1
			/* if the requested action is SWI_IRQ and it is called from within the interrupt-handler
			 * change the mode that will be restored after return from SWI to IRQ and disable IRQ */
			mrsls	r0, SPSR	/* use SPSR here, it is the value before entering SWI */
			andls	r0, r0, #(ARM_MODE_IRQ | ARM_F_BIT)
			orrls	r0, r0, #(ARM_MODE_IRQ | ARM_I_BIT)
			msrls	SPSR_c, r0	/* if returning from SWI, SPSR will be copied to CPSR resulting in mode change */
		ldmia	sp!, {r0,r1}		/* restore the saved r0 and r1 from the beginning. */
		movs	pc, lr			/* return and update CPSR */
SWI_Handler_a:
	/* SWI-Handler template for Assembler: */
/*
	ldr	r1, =SWI_
	cmp	r0, r1
	bne	SWI_Handler_b
		TODO: Code for Handler goes here:
SWI_Handler_b:
*/
	/* when all Assembler actions are tested, call the C-Handler.
	 */
	/* call the C-Routine for SWI-Handling in SYSTEM mode!
	 * It uses the UserMode Stack, allowing usual C-Handling,
	 * and runs in privileged mode.
	 */
	/* switch to System Mode, leave interrupts as they are */
	mrs	r1, CPSR
	and	r1, r1, #(ARM_MODE_System | ARM_F_BIT | ARM_I_BIT)
	orr	r1, r1, #ARM_MODE_System
	msr	CPSR_c, r1

	/* save registers in System(User)Mode stack */
	stmfd	sp!, {r2-r12,lr}

	/* the C-Routine *must not* use the SWI-Instruction! */
	/* the parameter (SWI-value) is still in r0 */
	.extern SWI_CHandler
	CALL_FUNC	SWI_CHandler

	/* restore registers from user mode stack */
	ldmia	sp!, {r2-r12,lr}

	/* switch back to the supervisor state */
	mrs	r0, CPSR
	and	r0, r0, #(ARM_MODE_Supervisor | ARM_F_BIT | ARM_I_BIT)
	orr	r0, r0, #ARM_MODE_Supervisor
	msr	CPSR_c, r0

	/* restore registers from supervisor stack */
	ldmia	sp!, {r0,r1}
	/* return from SWI */
	movs	pc, lr
#else
	/* if we do not need SWI, just return here without doing anything... */
	movs	pc, lr
#endif

.size SWI_Handler, . - SWI_Handler
.endfunc
